#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import sqlite3
import argparse
import time
#import line_profiler

cl_to_conflict = {}
new_id_to_old_id = {}
class Query:
    def __init__(self, dbfname):
        self.conn = sqlite3.connect(dbfname)
        self.c = self.conn.cursor()
        self.cl_used = []
        self.cl_used_num = 0
        self.cl_used_total = 0
        self.children_set = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.commit()
        self.conn.close()

    def delete_tbls(self, table):
        queries = """
DROP TABLE IF EXISTS `{table}`;
create table `{table}` ( `clauseID` bigint(20) NOT NULL, `used_at` bigint(20) NOT NULL, `weight` float NOT NULL);
""".format(table=table)

        for l in queries.split('\n'):
            t2 = time.time()

            if opts.verbose:
                print("Running query: ", l)
            self.c.execute(l)
            if opts.verbose:
                print("Query T: %-3.2f s" % (time.time() - t2))

    def get_conflicts(self):
        query= """select id, conflicts from set_id_confl;"""
        self.c.execute(query)
        while True:
            n = self.c.fetchone()
            if n is None:
                break

            ID = n[0]
            confl = n[1]
            if ID in cl_to_conflict:
                print("ERROR: ID %d in cl_to_conflict already. Old value: %d new value: %d" % (ID, cl_to_conflict[ID], confl))
                print("Likely, the issue is with `reloc` in FRAT, which is used in varreplacer. That is NOT understood by this system");
                exit(-1)
            cl_to_conflict[ID] = confl

        if opts.verbose:
            print("Got ID-conflict data")

    def get_updates(self):
        query= """select old_id, new_id from update_id order by old_id;"""
        self.c.execute(query)
        while True:
            n = self.c.fetchone()
            if n is None:
                break

            old_id = n[0]
            new_id = n[1]
            if old_id in new_id_to_old_id:
                real_old_id = new_id_to_old_id[old_id][0]
            else:
                real_old_id = old_id

            new_id_to_old_id[new_id] = [real_old_id, 1.0, None]

        if opts.verbose:
            print("Got ID-update data")

    def dump_used_clauses(self):
        for table in ("used_clauses_anc", "used_clauses"):
            if table == "used_clauses":
                todump = filter(lambda c: c[1] == 1.0, self.cl_used)
            else:
                todump = self.cl_used

            self.c.executemany("""
            INSERT INTO %s (
            `clauseID`,
            `weight`,
            `used_at`)
            VALUES (?, ?, ?);""" % table, todump)

        self.cl_used = []
        self.cl_used_num = 0

    #@profile
    def deal_with_chain(self, line, final_resolvent_ID, tracked_already, confl):
        for chain_str in line:
            if chain_str[0] == '0':
                break
            chain_ID = int(chain_str)
            if opts.verbose:
                print("Cl ID %8d used for Cl ID %8d" % (chain_ID, final_resolvent_ID))
            if chain_ID in new_id_to_old_id:
                if opts.verbose:
                    print("--> tracked as %s" % new_id_to_old_id[chain_ID])
                data = new_id_to_old_id[chain_ID]
                if data[2] == None:
                    # this is an original parent
                    count_as_confl = confl
                    assert data[1] == 1.0
                else:
                    # this is a child
                    count_as_confl = data[2]
                    assert data[1] < 1.0

                new_item = [data[0], data[1], count_as_confl]
                self.cl_used.append(new_item)
                self.cl_used_num+=1
                self.cl_used_total+=1
                if opts.verbose:
                    print("--> USED: %s" % new_item)
                if self.cl_used_num > 10000:
                    self.dump_used_clauses()

                # This chain_ID is tracked. Let's track Clause "final_resolvent_ID", it will be a child
                # NOTE: we only track it as ONE of the children... which is not ideal.
                if tracked_already:
                    if opts.verbose:
                        print("-----> Can't track Cl ID %d as child, already tracked either as MAIN or as a child" % final_resolvent_ID)
                    continue

                # final_resolvent_ID is not tracked already. We'll track it according to its ancestor
                data = new_id_to_old_id[chain_ID]
                chain_ID_upd = data[0]
                val = 0.5*data[1]
                if data[2] is not None:
                    # This is a child of a child, so parent's conf needs to be bumped
                    anc_confl = data[2]
                else:
                    # Children of this will need confl to be bumped
                    anc_confl = confl

                # Don't bother if it's less than 0.05
                if val <= 0.05:
                    continue

                if opts.verbose:
                    print("-----> Therefore, we will track ID %d with val %f to count as ID %d, confl %d" % (final_resolvent_ID, val, chain_ID_upd, anc_confl))
                new_id_to_old_id[final_resolvent_ID] = [chain_ID_upd, val, anc_confl]
                tracked_already = True
                self.children_set += 1

    #@profile
    def fix_up_frat(self, fratfile):
        with open(fratfile, "r") as f:
            for line in f:
                line=line.strip()
                if len(line) == 0:
                    continue
                if line[0] != "a":
                    continue

                if opts.verbose:
                    print("-------------****----------------------")
                    print("New line: %s" % line)
                line = line.split()
                if len(line) < 3:
                    print("ERROR: Line contains 1 or 2 elements??? It needs a/o/d/l/t + at least ID")
                    exit(-1)

                final_resolvent_ID = int(line[1])
                line = line[2:]
                tracked_already = False

                if final_resolvent_ID not in new_id_to_old_id:
                    if opts.verbose:
                        print("MAIN: Cl ID %8d is not tracked" % final_resolvent_ID)
                else:
                    if opts.verbose:
                        print("MAIN: Cl ID %8d is tracked, it is: %s" % (final_resolvent_ID, new_id_to_old_id[final_resolvent_ID]))
                    tracked_already = True

                if final_resolvent_ID not in cl_to_conflict:
                    print("ERROR: ID %8d not in cl_to_conflict" % final_resolvent_ID)
                    exit(-1)

                confl = cl_to_conflict[final_resolvent_ID]
                if opts.verbose:
                    print("MAIN: Cl ID %8d is generated at confl %d" % (final_resolvent_ID, confl))

                found = False
                for i in range(len(line)):
                    if line[i] == "l":
                        line = line[i+1:]
                        found = True
                        break

                if not found:
                    #if opts.verbose:
                    print("No explanation on line %s for ID %d" % (line, final_resolvent_ID))
                    continue

                self.deal_with_chain(line, final_resolvent_ID, tracked_already, confl)


if __name__ == "__main__":
    usage = """usage: %(prog)s [opts] sqlite_db usedCls

Adds used_clauses to the SQLite database"""

    parser = argparse.ArgumentParser(usage=usage)


    parser.add_argument("fratfile", type=str, metavar='FRAT')
    parser.add_argument("sqlitedb", type=str, metavar='SQLITEDB')
    parser.add_argument("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")

    opts = parser.parse_args()

    if opts.sqlitedb is None:
        print("Error you must give an SQLite DB. Please follow usage.")
        print(usage)
        exit(-1)

    if opts.fratfile is None:
        print("Error you must give the minimized FRAT file.")
        print(usage)
        exit(-1)

    print("Using FRAT file %s" % opts.fratfile)
    print("Using sqlite3db file %s" % opts.sqlitedb)

    with Query(opts.sqlitedb) as q:
        q.delete_tbls("used_clauses")
        q.delete_tbls("used_clauses_anc")
        q.get_conflicts()
        q.get_updates()
        q.fix_up_frat(opts.fratfile)
        # dump remaining ones (we dump 1000-by-1000)
        q.dump_used_clauses()

        print("Total num:    %10d" % q.cl_used_total)
        print("Children set: %10d" % q.children_set)

    exit(0)



#!/usr/bin/env python

import sys
of = sys.argv[1]
f = open(of, "r")
text = f.read()
mylines = text.splitlines()
i = 0;
time = []
for line in mylines :
  #print("line:", line)
  time.append(float(line.split()[0]))
  #print "t:%f" %(time[i])
  i += 1

lastnum = -1
for a in range(0, 4999, 1):
  num = 0
  for t in time:
    #print "t: %f a: %d" %(t, a)
    if (t < a) :
      num += 1

  if (lastnum != num) :
      print("%d \t%d" %(num, a))
  lastnum = num

f.close();
exit(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import with_statement  # Required in 2.5
from __future__ import print_function
import sys
import subprocess


def query_yes_no(question, default="no"):
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


num = 18
ignore = "1,2,8,9,10,11,15,5,14,13,3"
files = "/home/soos/media/sat/out/new/out-reconf-6776906.wlm01-*/*.out"

# for testing
#num = 8
#ignore = "0,1,2,3,5,6"
#files = "/home/soos/media/sat/out/new2/out-reconf-6776906.wlm01-*/*.out"

ignore_elems = {}
for x in ignore.split(","):
    x = x.strip()
    if x == "":
        continue

    x = int(x)
    ignore_elems[x] = True

subprocess.call("rm outs/*", shell=True)
toexec = "./reconf.py -n %d -i %s  -f outs/out  %s" % (num, ignore, files)
f = open("output", "w")
ret = subprocess.call(toexec, shell=True, stdout=f)
f.close()
if ret != 0:
    print("ERROR: reconf call exited non-zero: %s" % toexec)
    exit(-1)

for i in range(num):
    if i in ignore_elems:
        continue

    print("reconf with %d" % i)
    subprocess.call("cp outs/reconf.names outs/out%d.names" % i, shell=True)
    subprocess.call("c5.0 -u 20 -f outs/out%d -r > outs/out%d.c50.out" % (i, i), shell=True)

subprocess.call("./tocpp.py -i %s -n %d > ../../src/satzilla_features_to_reconf.cpp" % (ignore, num),
                shell=True)

subprocess.call("sed -i 's/red-/red_cl_distrib./g' ../../src/satzilla_features_to_reconf.cpp",
                shell=True)

upload = query_yes_no("Upload to AWS?")
if upload:
    subprocess.call("aws s3 cp ../../src/satzilla_features_to_reconf.cpp s3://msoos-solve-data/solvers/", shell=True)
    print("Uploded to AWS")
else:
    print("Not uploaded to AWS")





#
# CryptoMiniSat
#
# Copyright (c) 2009-2017, Mate Soos. All rights reserved.
# Copyright (c) 2017, Pierre Vignet
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import sys
import os
from setuptools import Extension, setup
import sysconfig
import toml
import pathlib
from sys import platform

def _parse_toml(pyproject_path):
    pyproject_text = pyproject_path.read_text()
    pyproject_data = toml.loads(pyproject_text)
    return pyproject_data['project']['version']

picosatlib = ('picosatlib', {
    'sources': [
               "src/picosat/picosat.c",
               "src/picosat/version.c"],
    'language' : "c",
    'define_macros' : [("TRACE", "ON")],
    'include_dirs' : ["src/picosat/"]
    })


def gen_modules(version):

    if platform == "win32" or platform == "cygwin":
        extra_compile_args_val = ['-I../', '-Isrc/', '/std:c++17', "/DCMS_FULL_VERSION=\""+version+"\""]
        define_macros_val = [("TRACE", "")]

    else:
        extra_compile_args_val = ['-I../', '-Isrc/', '-std=c++17']
        define_macros_val = [("TRACE", ""), ("CMS_FULL_VERSION", "\""+version+"\"")]

    modules = Extension(
        name = "pycryptosat",
        include_dirs = ["src/"],
        sources = ["python/src/pycryptosat.cpp",
                   "python/src/GitSHA1.cpp",
                   "src/bva.cpp",
                   "src/cardfinder.cpp",
                   "src/ccnr_cms.cpp",
                   "src/ccnr.cpp",
                   "src/clauseallocator.cpp",
                   "src/clausecleaner.cpp",
                   "src/cnf.cpp",
                   "src/completedetachreattacher.cpp",
                   "src/cryptominisat_c.cpp",
                   "src/cryptominisat.cpp",
                   "src/datasync.cpp",
                   "src/distillerbin.cpp",
                   "src/distillerlitrem.cpp",
                   "src/distillerlong.cpp",
                   "src/distillerlongwithimpl.cpp",
                   "src/frat.cpp",
                   "src/gatefinder.cpp",
                   "src/gaussian.cpp",
                   "src/get_clause_query.cpp",
                   "src/hyperengine.cpp",
                   "src/intree.cpp",
                   "src/lucky.cpp",
                   "src/matrixfinder.cpp",
                   "src/occsimplifier.cpp",
                   "src/packedrow.cpp",
                   "src/propengine.cpp",
                   "src/reducedb.cpp",
                   "src/sccfinder.cpp",
                   "src/searcher.cpp",
                   "src/searchstats.cpp",
                   "src/sls.cpp",
                   "src/solutionextender.cpp",
                   "src/solverconf.cpp",
                   "src/solver.cpp",
                   "src/str_impl_w_impl.cpp",
                   "src/subsumeimplicit.cpp",
                   "src/subsumestrengthen.cpp",
                   "src/varreplacer.cpp",
                   "src/xorfinder.cpp",
                   "src/oracle/oracle.cpp",
               ],
        extra_compile_args = extra_compile_args_val,
        define_macros=define_macros_val,
        language = "c++",
    )
    return modules

if __name__ == '__main__':
    pyproject_path = pathlib.Path('pyproject.toml')
    version = _parse_toml(pyproject_path)
    modules = gen_modules(version)
    setup(
        ext_modules =  [modules],
        libraries = [picosatlib],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018, Marcel Bargull, Mate Soos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from array import array
from collections import defaultdict
from itertools import chain
from time import time
import pycryptosat
import pycosat


def cms_setup(clauses, m):
    solver = pycryptosat.Solver()
    solver.add_clauses(clauses, max_var=m)
    return solver


def cms_solve(solver):
    return solver.solve()


def ps_setup(clauses, m):
    return pycosat.itersolve(clauses, vars=m)


def ps_solve(iter_sol):
    return next(iter_sol)


if __name__ == "__main__":
    n = 550*1000
    clause_list = list(chain.from_iterable(
        [(-x, x + n, x + int(n/2)), (-x, x + n, x + int(n/2))] for x in range(1, n+1)
    ))
    clause_array = array("i", chain.from_iterable(c + (0,) for c in clause_list))
    m = max(clause_array)

    all_times = defaultdict(list)
    my_list = ("CMS", cms_setup, cms_solve), ("Pycosat", ps_setup, ps_solve)
    #my_list = ("CMS", cms_setup, cms_solve), (None, None, None)
    for solver_type, setup, solve in list(my_list):
        if solver_type is None:
            continue

        my_type_list = ("list", clause_list), ("array", clause_array)
        #my_type_list = ("list", clause_list), (None, None)
        for clauses_type, clauses in (my_type_list):
            if clauses is None:
                continue

            for _ in range(5):
                t = time()
                solver = setup(clauses, m)
                all_times["setup {:<7} {:<5}".format(solver_type, clauses_type)].append(time() - t)
                t = time()
                solve(solver)
                all_times["solve {:<7} {:<5}".format(solver_type, clauses_type)].append(time() - t)
    for text, times in sorted(all_times.items()):
        print(text, *map("{:5.3f}".format, sorted(times)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA

from __future__ import with_statement  # Required in 2.5
from __future__ import print_function

from verifier import *
import unittest


class Map(dict):
    """
    Example:
    m = Map({'first_name': 'Eduardo'}, last_name='Pool', age=24, sports=['Soccer'])
    """
    def __init__(self, *args, **kwargs):
        super(Map, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.iteritems():
                self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)


class TestVerifier(unittest.TestCase):
    def setUp(self):
        options = Map({"verbose": False, "maxtime": 60, "maxtimediff": 160})
        self.s = solution_parser(options)

    def test_sat_cl(self):
        self.assertTrue(self.s._check_regular_clause("1 2 3", {1: True}))
        self.assertTrue(self.s._check_regular_clause("1 2 3", {2: True}))
        self.assertTrue(self.s._check_regular_clause("1 2 -3", {3: False}))

    def test_unsat_cl(self):
        self.assertRaises(NameError, self.s._check_regular_clause, "-1 2 3", {1: True})
        self.assertRaises(NameError, self.s._check_regular_clause, "-1 2 3", {})
        self.assertRaises(NameError, self.s._check_regular_clause, "-1 2 3 0", {})
        self.assertRaises(NameError, self.s._check_regular_clause, "1 2 -3 0", {0: True})
        self.assertRaises(NameError, self.s._check_regular_clause, "1 2 3", {1: False, 2: False, 3: False})
        self.assertRaises(NameError, self.s._check_regular_clause, "-1 -2 -3", {1: True, 2: True, 3: True})

    def test_sat_xcl(self):
        self.assertTrue(NameError, self.s._check_xor_clause("x1 2 3", {1: True, 2: False, 3: False}))
        self.assertTrue(NameError, self.s._check_xor_clause("x1 2 3 0", {1: True, 2: False, 3: False}))

    def test_unsat_xcl(self):
        self.assertRaises(NameError, self.s._check_xor_clause, "x1 2 3", {1: True})
        self.assertRaises(NameError, self.s._check_xor_clause, "x1 2 3", {1: False, 2: False, 3: False})

    def test_sol_parse_sat(self):
        unsat, s, _ = self.s.parse_solution_from_output(["s SAT", "v 1 2 3 0"])
        self.assertEqual(s, {1: True, 2: True, 3: True})

    def test_sol_parse_unsat(self):
        unsat, _, _ = self.s.parse_solution_from_output(["s UNSAT\n"])
        self.assertTrue(unsat)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.


from __future__ import with_statement  # Required in 2.5
from __future__ import print_function
import sys
import random

if len(sys.argv) !=4:
    print("ERROR: You must give 3 arguments: vars, Numxors and K")
    exit(-1)

vs = int(sys.argv[1])
numxors = int(sys.argv[2])
k = int(sys.argv[3])

print("c Will generate %d variable, %d XORs and then all variables's values sum will be at least %d" % (vs, numxors, k))

for i in range(numxors):
    thisxor = []
    for x in range(vs):
        if random.choice([True, False]):
            thisxor.append(x)
    print("c XOR:", thisxor)

    out = "x"
    if random.choice([True, False]):
        out += "-"
    for x in thisxor:
        out += "%d " % (x+1)
    out+="0\nc "
    print(out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import with_statement  # Required in 2.5
from __future__ import print_function
import subprocess
import os
import sys
import time
import random
from random import choice
import optparse
import glob
import resource
from verifier import *
from functools import partial

print("our CWD is: %s files here: %s" % (os.getcwd(), glob.glob("*")))
sys.path.append(os.getcwd())
print("our sys.path is", sys.path)

from debuglib import *


class PlainHelpFormatter(optparse.IndentedHelpFormatter):

    def format_description(self, description):
        if description:
            return description + "\n"
        else:
            return ""


usage = "usage: %prog [options] --fuzz/--regtest/--checkdir/filetocheck"
desc = """Fuzz the solver with fuzz-generator: ./fuzz_test.py
"""


def set_up_parser():
    parser = optparse.OptionParser(usage=usage, description=desc,
                                   formatter=PlainHelpFormatter())
    parser.add_option("--exec", metavar="SOLVER", dest="solver",
                      default="../../build/cryptominisat5",
                      help="SAT solver executable. Default: %default")

    parser.add_option("--extraopts", "-e", metavar="OPTS",
                      dest="extra_options", default="",
                      help="Extra options to give to SAT solver")

    parser.add_option("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")

    # for fuzz-testing
    parser.add_option("--seed", dest="fuzz_seed_start",
                      help="Fuzz test start seed. Otherwise, random seed is picked"
                      " (printed to console)", type=int)

    parser.add_option("--fuzzlim", dest="fuzz_test_lim", type=int,
                      help="Number of fuzz tests to run"
                      )
    parser.add_option("--dovalgrind", dest="novalgrind", default=True,
                      action="store_false", help="Use valgrind installed -- NOT recommended, compile with SANTIZE instead!")

    parser.add_option("--valgrindfreq", dest="valgrind_freq", type=int,
                      default=10, help="1 out of X times valgrind will be used. Default: %default in 1")

    parser.add_option("--gauss", dest="gauss", default=False,
                      action="store_true",
                      help="Concentrate fuzzing gauss")
    parser.add_option("--sampling", dest="only_sampling", default=False,
                      action="store_true",
                      help="Concentrate fuzzing sampling variables")
    parser.add_option("--maxth", "-m", dest="max_threads", default=20,
                      type=int, help="Max number of threads")

    parser.add_option("--tout", "-t", dest="maxtime", type=int, default=25,
                      help="Max time to run. Default: %default")

    parser.add_option("--textra", dest="maxtimediff", type=int, default=5,
                      help="Extra time on top of timeout for processing."
                      " Default: %default")

    return parser


def fuzzer_call_failed(fname):
    print("OOps, fuzzer executable call failed!")
    print("Did you build with cmake -DENABLE_TESTING=ON? Did you do git submodules init & update?")
    print("Here is the output:")

    print("**** ----- ****")
    with open(fname, "r") as a:
        for x in a:
            print(x.strip())
    print("**** ----- ****")
    exit(-1)


class create_fuzz:

    def call_from_fuzzer(self, fuzzer, fname):
        seed = random.randint(0, 1000000)
        if len(fuzzer) == 1:
            call = "{0} {1} > {2}".format(fuzzer[0], seed, fname)
        elif len(fuzzer) == 2:
            call = "{0} {1} {2} > {3}".format(
                fuzzer[0], fuzzer[1], seed, fname)
        elif len(fuzzer) == 3:
            hashbits = (random.getrandbits(20) % 80) + 1
            call = "%s %s %d %s %d > %s" % (
                fuzzer[0], fuzzer[1], hashbits, fuzzer[2], seed, fname)
        else:
            assert False, "Fuzzer must have at most 2 arguments"

        return call

    def create_fuzz_file(self, fuzzer, fuzzers, fname):
        # handle special fuzzer
        fnames_multi = []
        if len(fuzzer) == 2 and fuzzer[1] == "special":

            # sometimes just fuzz with all SAT problems
            fixed = random.getrandbits(1) == 1

            for _ in range(random.randrange(2, 4)):
                fname2 = unique_file("fuzzTest-multipart")
                fnames_multi.append(fname2)

                # chose a ranom fuzzer, not multipart
                fuzzer2 = ["multipart.py", "special"]
                while os.path.basename(fuzzer2[0]) == "multipart.py":
                    fuzzer2 = choice(fuzzers)

                # sometimes fuzz with SAT problems only
                if (fixed):
                    fuzzer2 = fuzzers[0]

                print("fuzzer2 used: %s" % fuzzer2)
                call = self.call_from_fuzzer(fuzzer2, fname2)
                print("calling sub-fuzzer: %s" % call)
                status = subprocess.call(call, shell=True)
                if status != 0:
                    fuzzer_call_failed(fname2)

            # construct multi-fuzzer call
            call = ""
            call += fuzzer[0]
            call += " "
            for name in fnames_multi:
                call += " " + name
            call += " > " + fname

            return call, fnames_multi

        # handle normal fuzzer
        else:
            return self.call_from_fuzzer(fuzzer, fname), []


def file_exists(fname):
    try:
        with open(fname):
            return True
    except IOError:
        return False


def print_version():
    command = options.solver + " --version"
    if options.verbose:
        print("Executing: %s" % command)
    p = subprocess.Popen(command.rsplit(), stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE, universal_newlines=True)

    consoleOutput, err = p.communicate()
    print("Version values: %s" % consoleOutput.strip())


class Tester:

    def __init__(self):
        self.ignoreNoSolution = False
        self.extra_opts_supported = self.list_options_if_supported(
            ["xor", "autodisablegauss", "sql", "breakid", "predshort"])
        self.sol_parser = solution_parser(options)
        self.sqlitedbfname = None
        self.only_sampling = False
        self.sampling_vars = []
        self.this_gauss_on = False
        self.num_threads = 1
        self.novalgrind = options.novalgrind

    def list_options_if_supported(self, tocheck):
        ret = []
        for elem in tocheck:
            if self.option_supported(elem):
                ret.append(elem)

        return ret

    def option_supported(self, option_name):
        command = options.solver
        command += " --hhelp"
        p = subprocess.Popen(
            command.rsplit(), stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
            universal_newlines=True)

        consoleOutput, err = p.communicate()

        for l in consoleOutput.split("\n"):
            tmp_option_name = "--" + option_name
            if tmp_option_name in l:
                return True

        return False

    def create_rnd_sched(self, string_list):
        opts = string_list.split(",")
        opts = [a.strip(" ") for a in opts]
        opts = sorted(list(set(opts)))
        if options.verbose:
            print("available schedule options: %s" % opts)

        sched = []
        for _ in range(int(random.gammavariate(12, 0.7))):
            sched.append(random.choice(opts))

        # just so that XOR is really found and used, so we can fuzz it
        if "autodisablegauss" in self.extra_opts_supported:
            sched.append("occ-xor")

        return sched

    def rnd_schedule_all(self):
        sched_opts = "scc-vrepl,"
        sched_opts += "sub-impl, intree-probe,"
        sched_opts += "sub-str-cls-with-bin, distill-cls, scc-vrepl, sub-impl,"
        sched_opts += "sub-cls-with-bin,"
        sched_opts += "str-impl, sub-str-cls-with-bin, distill-cls, scc-vrepl,"
        sched_opts += "occ-backw-sub-str, occ-backw-sub, occ-xor, occ-clean-implicit, occ-bve, occ-bva,"
        sched_opts += "renumber, must-renumber,"
        sched_opts += "card-find, breakid,"
        sched_opts += "occ-lit-rem, distill-bins, occ-resolv-subs,occ-rem-with-orgates"

        # type of schedule
        cmd = ""
        sched = self.create_rnd_sched(sched_opts)
        sched = ",".join(sched)

        if sched != "":
            cmd += "--schedule %s " % sched

        sched = ",".join(self.create_rnd_sched(sched_opts))
        if sched != "":
            cmd += "--preschedule %s " % sched

        return cmd

    def random_options(self):
        self.sqlitedbfname = None
        cmd = " --zero-exit-status "

        # disable gauss when gauss is compiled in but asked not to be used
        #if not self.this_gauss_on and "autodisablegauss" in self.extra_opts_supported:
            #cmd += "--gauss 0 "

        cmd += "--presimp %d " % random.choice([1]*10+[0])
        if not options.gauss:
            cmd += "--confbtwsimp %d " % random.choice([100, 1000])
            cmd += "--nextm %f " % random.choice([0.1, 0.01, 0.001])
            cmd += "--everylev1 %d " % random.choice([122, 1222, 12222])
            cmd += "--everylev2 %d " % random.choice([133, 1333, 14444])

        if "breakid" in self.extra_opts_supported:
            cmd += "--breakid %d " % random.choice([1]*10+[0])
            cmd += "--breakideveryn %d " % random.choice([1]*10+[3])
            cmd += "--breakidcls %d " % random.choice([0, 1, 2, 3, 10]+[50]*4)
            cmd += "--breakidtime %d " % random.choice([10000]*5+[1])

        if options.gauss:
            sls = 0
            cmd += "--autodisablegauss %s " % random.choice([0]*15+[1])

            # "Maximum number of matrixes to treat.")
            cmd += "--maxnummatrices %s " % int(random.gammavariate(1.5, 20.0))

        # SLS
        cmd += "--sls %d " % random.choice([0, 1])
        cmd += "--slsgetphase %d " % random.choice([0, 0, 0, 1])
        cmd += "--yalsatmems %d " % random.choice([1, 2, 5])
        cmd += "--walksatruns %d " % random.choice([2, 15, 20])

        # polarities
        cmd += "--polar %s " % random.choice(["true", "false", "rnd", "auto"])

        cmd += "--mustrenumber %d " % random.choice([0, 1])
        cmd += "--diffdeclevelchrono %d " % random.choice([1, random.randint(1, 1000), -1])
        cmd += "--bva %d " % random.choice([1, 1, 1, 0])
        cmd += "--bvaeveryn %d " % random.choice([1, random.randint(1, 20)])

        if self.only_sampling:
            cmd += "--onlysampling "
            cmd += "--sampling "
            cmd += ",".join(["%s" % x for x in self.sampling_vars]) + " "

        if random.choice([True, False]):
            cmd += "--varsperxorcut %d " % random.randint(4, 6)
            cmd += "--tern %d " % random.choice([0, 1])
            cmd += "--terntimelim %d " % random.choice([1, 10, 100])
            cmd += "--ternkeep %d " % random.choice([0, 0.001, 0.5, 300])
            cmd += "--terncreate %d " % random.choice([0, 0.001, 0.5, 300])
            cmd += "--locgmult %.12f " % random.gammavariate(0.5, 0.7)
            cmd += "--varelimover %d " % random.gammavariate(1, 20)
            cmd += "--memoutmult %0.12f " % random.gammavariate(0.05, 10)
            cmd += "--verb %d " % random.choice([0, 0, 0, 0, 1, 2])
            cmd += "--detachxor %d " % random.choice([0, 1, 1, 1, 1])
            cmd += "--restart %s " % random.choice(
                ["geom", "glue", "luby"])
            cmd += "--adjustglue %f " % random.choice([0, 0.5, 0.7, 1.0])
            cmd += "--gluehist %s " % random.randint(1, 500)
            cmd += "--updateglueonanalysis %s " % random.randint(0, 1)
            # cmd += "--clean %s " % random.choice(["size", "glue", "activity",
            # "prconf"])
            cmd += "--occredmax %s " % random.randint(0, 100)
            cmd += "--distill %s " % random.randint(0, 1)
            cmd += "--recur %s " % random.randint(0, 1)
            cmd += "--varelimcheckres %s " % random.randint(0, 1)
            cmd += "--implicitmanip %s " % random.randint(0, 1)
            cmd += "--occsimp %s " % random.randint(0, 1)
            cmd += "--occirredmaxmb %s " % random.gammavariate(0.2, 5)
            cmd += "--occredmaxmb %s " % random.gammavariate(0.2, 5)
            cmd += "--implsubsto %s " % random.choice([0, 10, 1000])
            cmd += "--sync %d " % random.choice([100, 1000, 6000, 100000])
            cmd += "-m %0.12f " % random.gammavariate(0.1, 5.0)
            cmd += "--maxsccdepth %d " % random.choice([0, 1, 100, 100000])

            # more more minim
            cmd += "--moremoreminim %d " % random.choice([1, 1, 1, 0])
            cmd += "--moremorealways %d " % random.choice([1, 1, 1, 0])

            if self.this_gauss_on:
                if random.randint(0,1000) < 10:
                    # Set maximum no. of rows for gaussian matrix."
                    cmd += "--maxmatrixrows %s " % int(random.gammavariate(1, 15.0))

                    # "Set minimum no. of rows for gaussian matrix.
                    cmd += "--minmatrixrows %s " % int(random.gammavariate(1, 15.0))

            if "sql" in self.extra_opts_supported and random.randint(0, 3) > 0 and self.num_threads == 1:
                cmd += "--sql 2 "
                self.sqlitedbfname = unique_file("fuzz", ".sqlitedb")
                cmd += "--sqlitedb %s " % self.sqlitedbfname
                cmd += "--sqlitedboverwrite 1 "
                cmd += "--cldatadumpratio %0.3f " % random.choice([0.9, 0.1, 0.7])

        # the most buggy ones, don't turn them off much, please
        if random.choice([1, 0, 0]):
            opts = ["scc", "varelim", "strengthen", "intree",
                    "renumber", "savemem", "moreminim", "gates",
                    "schedsimp", "otfhyper"]

            if "xor" in self.extra_opts_supported:
                opts.append("xor")

            for opt in opts:
                if opt == "xor" and self.this_gauss_on:
                    continue

                cmd += "--%s %d " % (opt, random.choice([0, 1, 1, 1, 1]))

            cmd += self.rnd_schedule_all()

        return cmd

    def execute(self, fname, fname_frat=None, fixed_opts="", rnd_opts=None):
        if os.path.isfile(options.solver) is not True:
            print("Error: Cannot find CryptoMiniSat executable.Searched in: '%s'" %
                  options.solver)
            print("Error code 300")
            exit(300)

        for f in glob.glob("%s-debugLibPart*.output" % fname):
            os.unlink(f)

        # construct command
        command = ""
        if not self.novalgrind and random.randint(1, options.valgrind_freq) == 1:
            command += "valgrind -q --leak-check=full  --error-exitcode=9 "
        command += options.solver
        if rnd_opts is None:
            rnd_opts = self.random_options()
        command += rnd_opts
        if self.needDebugLib:
            command += "--debuglib %s " % fname
        command += "--threads %d " % self.num_threads
        command += options.extra_options + " "
        command += fixed_opts + " "
        if fname is not None:
            command += " %s " % fname
        if fname_frat:
            command += " %s " % fname_frat

        if options.verbose:
            print("Executing before doalarm/pipes: %s " % command)

        # print time limit
        if options.verbose:
            print("CPU limit of parent (pid %d)" % os.getpid(), resource.getrlimit(resource.RLIMIT_CPU))

        # if need time limit, then limit
        err_fname = unique_file("err", ".out")
        out_fname = unique_file("out", ".out")

        if self.num_threads == 1:
            time_limit = partial(setlimits, options.maxtime)
        else:
            time_limit = None

        myalarm = "./doalarm -t real %d "% options.maxtime
        toexec = myalarm + command + " 1>" + out_fname + " 2> " + err_fname
        print("Executing: " + toexec)
        retcode = os.system(toexec)

        # print time limit after child startup
        #if options.verbose:
            #print("CPU limit of parent (pid %d) after startup of child: %s secs" %
                  #(os.getpid(), resource.getrlimit(resource.RLIMIT_CPU)))

        # Get solver output
        consoleOutput = ""
        with open(out_fname, "r") as f:
            for line in f:
                consoleOutput += line
        #retcode = p.returncode
        #err_file.close()
        with open(err_fname, "r") as err_file:
            found_something = False
            for line in err_file:
                line = line.strip()
                if line == "":
                    continue

                if options.verbose:
                    print("ERR line: ", line)

                # let's not care about leaks for the moment
                if "LeakSanitizer: detected memory leaks" in line:
                    break

                # don't error out on issues related to UBSAN/ASAN
                # of clang of other projects
                noprob = ["std::_Ios_Fmtflags", "mzd.h", "lexical_cast.hpp",
                      "MersenneTwister.h", "boost::any::holder", "~~~~~",
                      "SUMMARY", "process memory map", "==========="]

                ok = False
                for x in noprob:
                    if x in line:
                        ok = True

                if not ok:
                    found_something = True
                    errline = line

            if found_something:
                print("Error line while executing: %s" % errline.strip())
                exit(-1)

        os.unlink(err_fname)
        if self.sqlitedbfname is not None:
            os.unlink(self.sqlitedbfname)

        if options.verbose:
            print("CPU limit of parent (pid %d) after child finished executing: %s" %
                  (os.getpid(), resource.getrlimit(resource.RLIMIT_CPU)))

        os.unlink(out_fname)
        return consoleOutput, retcode

    def check(self, fname, fname_frat=None,
              checkAgainst=None,
              fixed_opts="",
              rnd_opts=None):

        consoleOutput = ""
        if checkAgainst is None:
            checkAgainst = fname
        curr_time = time.time()

        # Do we need to solve the problem, or is it already solved?
        consoleOutput, retcode = self.execute(
            fname, fname_frat=fname_frat,
            fixed_opts=fixed_opts, rnd_opts=rnd_opts)

        # if time was limited, we need to know if we were over the time limit
        # and that is why there is no solution
        diff_time = time.time() - curr_time
        if diff_time > (options.maxtime - options.maxtimediff):
            print("Too much time to solve, aborted!")
            return None

        print("Within time limit: %.2f s" % diff_time)
        print("filename: %s" % fname)

        if options.verbose:
            print(consoleOutput)

        if retcode != 0:
            print("Return code of CryptoMiniSat is not 0, it is: %d -- error!" % retcode)
            exit(-1)

        # if library debug is set, check it
        if (self.needDebugLib):
            must_check_unsat = True
            if options.gauss:
                must_check_unsat = random.choice([False]*15+[True])
            self.sol_parser.check_debug_lib(checkAgainst, must_check_unsat)

        print("Checking console output...")
        unsat, solution, _ = self.sol_parser.parse_solution_from_output(
            consoleOutput.split("\n"), self.ignoreNoSolution)

        if not unsat:
            if len(self.sampling_vars) != 0:
                self.sol_parser.sampling_vars_solution_check(fname, self.sampling_vars, solution)
            else:
                self.sol_parser.test_found_solution(solution, checkAgainst)

            return

        # it's UNSAT, let's check with FRAT
        if fname_frat:
            toexec = "./frat-rs elab {fname_frat} {cnf} -v"
            toexec = toexec.format(cnf=fname, fname_frat=fname_frat)
            print("Checking with FRAT.. ", toexec)
            p = subprocess.Popen(toexec.rsplit(),
                    stdout=subprocess.PIPE,
                    universal_newlines=True)

            consoleOutput2 = p.communicate()[0]
            diff_time = time.time() - curr_time

            # find verification code
            foundVerif = False
            fratLine = ""
            for line in consoleOutput2.split('\n'):
                if len(line) >= 8:
                    if line[0:8] == "VERIFIED":
                        fratLine = line
                        foundVerif = True

            # Check whether we have found a verification code
            if foundVerif is False:
                print("verifier error! It says: %s" % consoleOutput2)
                assert foundVerif, "Cannot find FRAT verification code!"
            else:
                print("OK, FRAT says: %s" % fratLine)

        if options.gauss:
            if random.choice([True, True, True, True, False]):
                print("NOT verifying with other solver, it's Gauss so too expensive")
                return

        # check with other solver
        ret = self.sol_parser.check_unsat(checkAgainst)
        if ret is None:
            print("Other solver time-outed, cannot check")
        elif ret is True:
            print("UNSAT verified by other solver")
        else:
            print("Grave bug: SAT-> UNSAT : Other solver found solution!!")
            exit()

    def fuzz_test_one(self):
        print("--- NORMAL TESTING ---")
        self.num_threads = random.choice([1]+[random.randint(2,4)])
        self.num_threads = min(options.max_threads, self.num_threads)
        self.this_gauss_on = "autodisablegauss" in self.extra_opts_supported

        # frat turns off a bunch of systems, like symmetry breaking so use it about 50% of time
        self.frat = self.num_threads == 1 and (random.randint(0, 10) < 5)

        self.sqlitedbfname = None
        self.only_sampling = random.choice([True, False, False, False, False]) and not self.frat

        if options.only_sampling:
            self.frat = False
            self.only_sampling = True

        if self.frat:
            fuzzers = fuzzers_frat
        elif options.gauss:
            fuzzers = fuzzers_xor
        else:
            fuzzers = fuzzers_nofrat
        fuzzer = random.choice(fuzzers)

        fname = unique_file("fuzzTest")
        fname_frat = None
        if self.frat:
            fname_frat = unique_file("fuzzTest-frat")

        # create the fuzz file
        cf = create_fuzz()
        call, todel = cf.create_fuzz_file(fuzzer, fuzzers, fname)
        print("calling %s" % call)
        status = subprocess.call(call, shell=True)
        if status != 0:
            fuzzer_call_failed(fname)

        shuf_seed = random.randint(1, 100000)
        fname_shuffled = unique_file("fuzzTest")
        print("calling ./shuffle.py %s %s %s" % (fname, fname_shuffled, shuf_seed))
        shuffle_cnf(fname, fname_shuffled, shuf_seed)
        os.unlink(fname)
        fname = fname_shuffled

        if not self.frat and not self.only_sampling:
            print("->Multipart test")
            self.needDebugLib = True
            interspersed_fname = unique_file("fuzzTest-interspersed")
            seed_for_inters = random.randint(0, 1000000)
            intersperse(fname, interspersed_fname, seed_for_inters)
            print("Interspersed: ./intersperse.py %s %s %d" % (fname,
                                                               interspersed_fname,
                                                               seed_for_inters))
            os.unlink(fname)
        else:
            self.needDebugLib = False
            interspersed_fname = fname

        # calculate sampling vars
        self.sampling_vars = []
        if self.only_sampling:
            max_vars = self.sol_parser.max_vars_in_file(fname)
            assert max_vars > 0

            self.sampling_vars = []
            myset = {}
            for _ in range(random.randint(1, 50)):
                x = random.randint(1, max_vars)
                if x not in myset:
                    self.sampling_vars.append(x)
                    myset[x] = 1

            # don't do it for 0-length sampling vars
            if len(self.sampling_vars) == 0:
                self.only_sampling = False

        self.check(fname=interspersed_fname, fname_frat=fname_frat)

        # remove temporary filenames
        os.unlink(interspersed_fname)
        if fname_frat:
            os.unlink(fname_frat)
        for name in todel:
            os.unlink(name)

    def delete_file_no_matter_what(self, fname):
        try:
            os.unlink(fname)
        except:
            pass

fuzzers_noxor = [
    ["../../build/tests/sha1-sat/sha1-gen --nocomment --attack preimage --rounds 20",
     "--hash-bits", "--seed"],
    ["../../build/tests/sha1-sat/sha1-gen --nocomment --attack preimage --zero "
        "--message-bits 400 --rounds 8 --hash-bits 60",
     "--seed"],
    # ["build/cnf-fuzz-nossum"],
    ["../../build/tests/cnf-utils/cnf-fuzz-biere"],
    ["../../build/tests/cnf-utils/cnf-fuzz-biere"],
    ["../../build/tests/cnf-utils/cnf-fuzz-biere"],
    ["../../build/tests/cnf-utils/cnf-fuzz-biere"],
    ["../../build/tests/cnf-utils/makewff -cnf 3 250 1080", "-seed"],
    ["../../build/tests/cnf-utils/sgen4 -unsat -n 50", "-s"],
    ["../../build/tests/cnf-utils//sgen4 -sat -n 50", "-s"],
    ["../../utils/cnf-utils/cnf-fuzz-brummayer.py", "-s"],
    ["../../utils/cnf-utils/cnf-fuzz-xor.py", "--seed"],
    ["../../utils/cnf-utils/cnf-fuzz-xor.py", "--seed"],
    ["../../utils/cnf-utils/cnf-fuzz-xor.py", "--seed"],
    ["../../utils/cnf-utils/cnf-fuzz-xor.py", "--seed"],
    ["../../utils/cnf-utils/multipart.py", "special"]
]

fuzzers_xor = [
    ["../../utils/cnf-utils/xortester.py --varsmin 40", "--seed"],
    ["../../utils/cnf-utils/xortester.py --varsmin 60", "--seed"],
    ["../../utils/cnf-utils/xortester.py --varsmin 80", "--seed"],
    ["../../utils/cnf-utils/xortester.py --varsmin 100", "--seed"],
    ["../../utils/cnf-utils/xortester.py --varsmin 200", "--seed"],
    ["../../utils/cnf-utils/spacer_test.py", "--seed"],
    ["../../build/tests/sha1-sat/sha1-gen --xor --attack preimage --rounds 21",
     "--hash-bits", "--seed"],
]


if __name__ == "__main__":
    if not os.path.isdir("out"):
        print("Directory for outputs, 'out' not present, creating it.")
        os.mkdir("out")

    # parse options
    parser = set_up_parser()
    (options, args) = parser.parse_args()
    if options.valgrind_freq <= 0:
        print("Valgrind Frequency must be at least 1")
        exit(-1)

    fuzzers_frat = fuzzers_noxor
    fuzzers_nofrat = fuzzers_noxor + fuzzers_xor

    print_version()
    tester = Tester()
    tester.needDebugLib = False
    num = 0
    rnd_seed = options.fuzz_seed_start
    if rnd_seed is None:
        rnd_seed = random.randint(0, 1000*1000*100)

    while True:
        toexec = "./fuzz_test.py --fuzzlim 1 --seed %d " % rnd_seed
        if not options.novalgrind:
            toexec += "--dovalgrind "
        if options.valgrind_freq:
            toexec += "--valgrindfreq %d " % options.valgrind_freq
        if options.gauss:
            toexec += "--gauss "
        if options.only_sampling:
            toexec += "--sampling "
        toexec += "-m %d " % options.max_threads
        toexec += "-t %d " % options.maxtime

        print("")
        print("")
        print("--> To re-create fuzz-test below: %s" % toexec)

        random.seed(rnd_seed)
        tester.fuzz_test_one()
        rnd_seed += 1
        num += 1
        if options.fuzz_test_lim is not None and num >= options.fuzz_test_lim:
            exit(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import sklearn.ensemble
import sklearn.tree
import sqlite3
import argparse
import time
import pickle
import re
import pandas as pd
import numpy as np
import os.path
import sys
import helper
import sklearn
import functools
ver = sklearn.__version__.split(".")
if int(ver[1]) < 20:
    from sklearn.cross_validation import train_test_split
else:
    from sklearn.model_selection import train_test_split


def add_computed_features(df):
    print("Original number of features:", len(list(df)))
    print("Adding computed features...")
    cols = list(df)
    divide = functools.partial(helper.divide, df=df, features=cols, verb=options.verbose)

    if not options.picktime_only:
        # create "during"
        for col in cols:
            if "_at_fintime" in col:
                during_name = col.replace("_at_fintime", "_during")
                at_picktime_name = col.replace("_at_fintime", "_at_picktime")
                at_fintime_name = col
                if options.verbose:
                    print("fintime name: ", at_fintime_name)
                    print("picktime name: ", at_picktime_name)
                df[during_name] = df[at_fintime_name]-df[at_picktime_name]

        # remove picktime & fintime, only use "during"
        cols = list(df)
        for c in cols:
            if "at_picktime" in c or "at_fintime" in c:
                del df[c]

    else:
        # remove everything to do with "clauses_below" and "at_fintime"
        cols = list(df)
        for c in cols:
            if "var_data_fintime" in c:
                del df[c]

    if False:
        # per-conflicts, per-decisions, per-lits
        divisors = [
            "var_data_picktime.sumConflicts_at_picktime"
            , "var_data_picktime.sumClLBD_at_picktime"
            , "var_data_picktime.sumClSize_at_picktime"
            , "var_data_picktime.sumConflictClauseLits_at_picktime"
            # neutral below
            , "var_data_picktime.dec_depth"
            # below during
            , "var_data_picktime.inside_conflict_clause_antecedents_at_picktime"
            , "var_data_picktime.sumDecisions_below_during"
            , "var_data_picktime.sumPropagations_below_during"
            , "var_data_picktime.sumConflicts_below_during"
            , "var_data_picktime.sumAntecedents_below_during"
            , "var_data_picktime.sumConflictClauseLits_below_during"
            , "var_data_picktime.sumAntecedentsLits_below_during"
            , "var_data_picktime.sumClSize_below_during"
            , "var_data_picktime.sumClLBD_below_during"
            ]

        cols = list(df)
        for col in cols:
            if "x." not in col and "var_data_use" not in col:
                for divisor in divisors:
                    divide(col, divisor)

    # divide var_dist by szfeat, all-by-all
    if False:
        for c in cols:
            if "szfeat" in c:
                for c2 in cols:
                    if "var_dist" in c2:
                        divide(c2, c)

    # divide during by during, all-by-all
    if True:
        for c in cols:
            if "during" in c:
                for c2 in cols:
                    if "during" in c2:
                        divide(c2, c)

    df["var_dist.num_irred_cls"] = df["var_dist.num_irred_long_cls"] + df["var_dist.num_irred_bin_cls"]
    df["var_dist.num_red_cls"] = df["var_dist.num_red_long_cls"] + df["var_dist.num_red_bin_cls"]

    for red in ["red", "irred"]:
        divide("var_dist.{red}_num_times_in_bin_clause".format(red=red), "var_dist.num_{red}_bin_cls".format(red=red))
        divide("var_dist.{red}_num_times_in_long_clause".format(red=red), "var_dist.num_{red}_long_cls".format(red=red))
        divide("var_dist.{red}_satisfies_cl".format(red=red), "var_dist.num_{red}_cls".format(red=red))
        divide("var_dist.{red}_tot_num_lit_of_bin_it_appears_in".format(red=red), "var_dist.num_{red}_bin_cls".format(red=red))
        divide("var_dist.{red}_tot_num_lit_of_long_cls_it_appears_in".format(red=red), "var_dist.num_{red}_long_cls".format(red=red))
        divide("var_dist.{red}_sum_var_act_of_cls".format(red=red), "var_dist.num_{red}_long_cls".format(red=red))
        divide("var_dist.{red}_satisfies_cl".format(red=red), "var_dist.num_{red}_cls".format(red=red))
        divide("var_dist.{red}_falsifies_cl".format(red=red), "var_dist.num_{red}_cls".format(red=red))
        divide("var_dist.{red}_tot_num_lit_of_long_cls_it_appears_in".format(red=red), "var_dist.num_{red}_long_cls".format(red=red))
        divide("var_dist.{red}_sum_var_act_of_cls".format(red=red), "var_dist.num_{red}_long_cls".format(red=red))
        divide("var_dist.{red}_satisfies_cl".format(red=red), "var_dist.num_{red}_cls".format(red=red))
        divide("var_dist.{red}_sum_var_act_of_cls".format(red=red), "var_dist.num_{red}_long_cls".format(red=red))

    divide("var_dist.tot_act_long_red_cls", "var_dist.num_red_long_cls")

    divide("var_data_picktime.inside_conflict_clause_antecedents_during_at_picktime",
           "var_data_picktime.sumAntecedentsLits_at_picktime")

    divide("var_data_picktime.sumAntecedentsLits_below_during",
           "var_data_picktime.sumAntecedentsLits_at_picktime")






    divide("var_data_picktime.inside_conflict_clause_during_at_picktime",
           "var_data_picktime.sumConflicts_at_picktime")
    divide("var_data_picktime.inside_conflict_clause_antecedents_at_picktime",
        "var_data_picktime.sumAntecedentsLits_at_picktime")

    divide("var_data_picktime.num_decided", "var_data_picktime.sumDecisions_at_picktime")
    divide("var_data_picktime.num_decided_pos", "var_data_picktime.sumDecisions_at_picktime")
    divide("var_data_picktime.num_decided", "var_data_picktime.num_decided_pos")

    divide("var_data_picktime.num_propagated", "var_data_picktime.sumPropagations_at_picktime")
    divide("var_data_picktime.num_propagated_pos", "var_data_picktime.sumPropagations_at_picktime")
    divide("var_data_picktime.num_propagated", "var_data_picktime.num_propagated_pos")


    for c in cols:
        if "var_dist.num_" in c:
            del df[c]
    del df["var_dist.num_irred_cls"]
    del df["var_dist.num_red_cls"]

    # we divide these and then delete
    xs = [
        "var_dist.red_num_times_in_long_clause",
        "var_data_picktime.inside_conflict_clause_at_picktime",
        "var_data_picktime.sumClSize_at_picktime",
        "var_data_picktime.sumClLBD_at_picktime",
        "var_data_picktime.sumAntecedents_at_picktime",
        "var_data_picktime.sumAntecedentsLits_at_picktime",
        "var_data_picktime.sumConflictClauseLits_at_picktime",
        "var_data_picktime.inside_conflict_clause_antecedents_at_picktime",
        "var_data_picktime.inside_conflict_clause_antecedents_during_at_picktime",
        "var_data_picktime.sumAntecedentsLits_below_during",
        "var_data_picktime.inside_conflict_clause_glue_at_picktime",
        "var_data_picktime.inside_conflict_clause_glue_during_at_picktime"
        ]
    ys = [
        "var_data_picktime.sumConflicts_at_picktime",
        "var_data_picktime.num_decided",
        "var_data_picktime.num_propagated"
    ]
    for x in xs:
        for y in ys:
            divide(x, y)
        del df[x]



    todel = [
        "var_data_picktime.latest_vardist_feature_calc",

        "var_dist.red_falsifies_cl",
        "var_dist.irred_falsifies_cl",
        "var_dist.red_satisfies_cl",
        "var_dist.irred_satisfies_cl",

        "var_data_picktime.dec_depth",
        "var_dist.red_num_times_in_bin_clause",
        "var_dist.irred_sum_var_act_of_cls",
        "var_dist.tot_act_long_red_cls",
        "var_dist.red_tot_num_lit_of_long_cls_it_appears_in",
        "var_dist.red_sum_var_act_of_cls",
        #"rst.branch_strategy",

        "var_data_picktime.num_decided",
        "var_data_picktime.num_decided_pos",
        "var_data_picktime.num_propagated",
        "var_data_picktime.num_propagated_pos",
        "var_data_picktime.sumPropagations_at_picktime",
        "var_data_picktime.sumDecisions_at_picktime",

        "var_data_picktime.inside_conflict_clause_during_at_picktime"
        ]
    for d in todel:
        del df[d]

    for c in cols:
        if "rst." in c and "strategy" not in c and "restart_type" not in c:
            if c in list(df):
                del df[c]

    print("Done adding computed features. New number of features: ", len(list(df)))


def rem_useless_features(df):
    col = list(df)
    for c in col:
        if "restart_type" in c or "szfeat" in c:
            del df[c]

    # remove hidden data
    del df["sum_cl_use.num_used"]
    pass


class Learner:
    def __init__(self, df, funcname, fname, cluster_no):
        self.df = df
        self.func_name = funcname
        self.fname = fname
        self.cluster_no = cluster_no

    def cut_into_chunks(self):
        df["x.class"] = pd.qcut(
            df["x.num_used"],
            q=options.quantiles,
            duplicates='drop',
            labels=False)

        df['x.class'] = df['x.class'].astype(str)

    @staticmethod
    def fix_feat_name(x):
        if "during" in x or "clauses_below" in x:
            x = x.replace("var_data_", "")

        return x

    def one_classifier(self, features, to_predict, final, write_code=False):
        print("-> Number of features  :", len(features))
        print("-> Number of datapoints:", self.df.shape)
        print("-> Predicting          :", to_predict)

        # get smaller part to work on
        # also, copy it so we don't get warning about setting a slice of a DF
        if options.only_pecr >= 0.98:
            df = self.df.copy()
        else:
            _, df_tmp = train_test_split(self.df, test_size=options.only_pecr)
            df = df_tmp.copy()
            print("-> Number of datapoints after applying '--only':", df.shape)

        if options.dump_csv:
            fname = "mydump.csv"
            print("Dumping CSV data to:", fname)
            df.to_csv(fname, index=False, columns=sorted(list(df)))

        if options.check_row_data:
            helper.check_too_large_or_nan_values(df, features+["x.class"])
            print("Checked, all good!")

        train, test = train_test_split(df, test_size=0.33)
        X_train = train[features]
        y_train = train[to_predict]

        t = time.time()
        clf = None
        if final:
            split_point = helper.calc_min_split_point(
                df, options.min_samples_split)
            clf = sklearn.tree.DecisionTreeClassifier(
                max_depth=options.tree_depth,
                min_samples_split=split_point)
        else:
            clf = sklearn.ensemble.RandomForestClassifier(
                n_estimators=1000,
                max_features="sqrt")

        del df
        clf.fit(X_train, y_train)
        print("Training finished. T: %-3.2f" % (time.time() - t))

        if not final:
            best_features = helper.print_feature_ranking(
                clf, X_train,
                top_num_features=options.top_num_features,
                features=features,
                plot=options.show)
        else:
            if options.dot is not None:
                if not options.final_is_tree:
                    print("ERROR: You cannot use the DOT function on non-trees")
                    exit(-1)

                helper.output_to_classical_dot(
                    clf, features,
                    fname=options.dot + "-" + self.func_name)

            if options.basedir and write_code:
                c = helper.CodeWriter(clf, features, self.func_name, self.fname, options.verbose)
                c.func_signature = """
                const Solver*    solver
                , const VarData& varData
                , uint64_t       sumConflicts_during
                , uint64_t       sumDecisions_during
                , uint64_t       sumPropagations_during
                , uint64_t       sumAntecedents_during
                , uint64_t       sumAntecedentsLits_during
                , uint64_t       sumConflictClauseLits_during
                , uint64_t       sumDecisionBasedCl_during
                , uint64_t       sumClLBD_during
                , uint64_t       sumClSize_during
                , uint64_t       clauses_below
                """
                c.func_call = """
                solver
                , varData
                , sumConflicts_during
                , sumDecisions_during
                , sumPropagations_during
                , sumAntecedents_during
                , sumAntecedentsLits_during
                , sumConflictClauseLits_during
                , sumDecisionBasedCl_during
                , sumClLBD_during
                , sumClSize_during
                , clauses_below
                """
                c.per_func_defines = """"""
                c.file_header = """
                #include "clause.h"
                #include "vardata.h"
                #include "solver.h"
                #include <cmath>

                namespace CMSat {
                """
                c.fix_feat_name = self.fix_feat_name
                c.clean_up()
                c.print_full_code()

        precision, recall, accuracy = helper.conf_matrixes(
            test, features, to_predict, clf,
            toprint="test", average="micro")
        helper.conf_matrixes(
            train, features, to_predict, clf,
            toprint="train", average="micro")

        # TODO do L1 regularization
        # TODO do principal component analysis

        if not final:
            return best_features
        else:
            return prec+recall+acc

    def learn(self):
        self.cut_into_chunks()
        features = list(self.df)
        features.remove("clust")
        features.remove("x.class")
        features.remove("x.num_used")

        if options.raw_data_plots:
            pd.options.display.mpl_style = "default"
            self.df.hist()
            self.df.boxplot()

        if not options.only_final:
            top_n_feats = self.one_classifier(features, "x.class", final=False)
            if options.show:
                plt.show()

            if options.get_best_topn_feats is not None:
                greedy_features = helper.calc_greedy_best_features(
                    top_n_feats, options.get_best_topn_feats,
                    self)

            return

        best_features = [
            'var_data_picktime.flipped_confs_ago',
            '(var_dist.tot_act_long_red_cls/var_dist.num_red_long_cls)',
            'var_data_picktime.conflicts_since_decided',
            '(var_data_picktime.num_decided/var_data_picktime.num_decided_pos)',
            '(var_dist.red_num_times_in_long_clause/var_data_picktime.num_propagated)',
            '(var_dist.red_num_times_in_long_clause/var_data_picktime.num_decided)',
            'var_data_picktime.rel_activity_at_picktime',
            '(var_data_picktime.sumAntecedentsLits_below_during/var_data_picktime.sumAntecedentsLits_at_picktime)',
            '(var_data_picktime.num_propagated/var_data_picktime.num_propagated_pos)',
            'var_data_picktime.conflicts_since_propagated',
            ]

        self.one_classifier(best_features, "x.class",
                            final=True,
                            write_code=True)

        if options.show:
            plt.show()


if __name__ == "__main__":
    usage = "usage: %(prog)s [options] file.sqlite"
    parser = argparse.ArgumentParser(usage=usage)

    # dataframe
    parser.add_argument("fname", type=str, metavar='PANDASFILE')
    parser.add_argument("--verbose", "-v", action="store_true", default=False,
                        dest="verbose", help="Print more output")
    parser.add_argument("--top", default=40, type=int, metavar="TOPN",
                        dest="top_num_features", help="Candidates are top N features for greedy selector")
    parser.add_argument("--printfeat", action="store_true", default=False,
                        dest="print_features", help="Print features")
    parser.add_argument("--check", action="store_true", default=False,
                        dest="check_row_data", help="Check row data for NaN or float overflow")
    parser.add_argument("--greedy", default=None, type=int, metavar="TOPN",
                        dest="get_best_topn_feats", help="Greedy Best K top features from the top N features given by '--top N'")
    parser.add_argument("--show", default=False, action="store_true",
                        dest="show", help="Show graphs")
    parser.add_argument("--final", default=False, action="store_true",
                        dest="only_final", help="Only generate final predictor")
    parser.add_argument("--rawplots", action="store_true", default=False,
                        dest="raw_data_plots", help="Display raw data plots")
    parser.add_argument("--dot", type=str, default=None,
                        dest="dot", help="Create DOT file")
    parser.add_argument("--basedir", type=str,
                        dest="basedir", help="The base directory of where the CryptoMiniSat source code is")
    parser.add_argument("--clust", default=False, action="store_true",
                        dest="use_clusters", help="Use clusters")
    parser.add_argument("--conf", default=0, type=int,
                        dest="conf_num", help="Which predict configuration this is")
    parser.add_argument("--picktimeonly", default=False, action="store_true",
                        dest="picktime_only", help="Only use and generate pictime data")

    # tree/forest options
    parser.add_argument("--depth", default=None, type=int,
                        dest="tree_depth", help="Depth of the tree to create")
    parser.add_argument("--split", default=0.01, type=float, metavar="RATIO",
                        dest="min_samples_split", help="Split in tree if this many samples or above. Used as a percentage of datapoints")
    parser.add_argument("--numtrees", default=5, type=int,
                        dest="num_trees", help="How many trees to generate for the forest")

    # data filtering
    parser.add_argument("--only", default=0.99, type=float,
                        dest="only_pecr", help="Only use this percentage of data")
    parser.add_argument("-q", default=2, type=int, metavar="QUANTS",
                        dest="quantiles", help="Number of quantiles we want")
    parser.add_argument("--nocomputed", default=False, action="store_true",
                        dest="no_computed", help="Don't add computed features")
    parser.add_argument("--csv", action="store_true", default=False,
                        dest="dump_csv", help="Dump CSV (for weka)")

    # type of classifier
    parser.add_argument("--tree", default=False, action="store_true",
                        dest="final_is_tree", help="Final classifier should be a tree")
    parser.add_argument("--svm", default=False, action="store_true",
                        dest="final_is_svm", help="Final classifier should be an svm")
    parser.add_argument("--lreg", default=False, action="store_true",
                        dest="final_is_logreg", help="Final classifier should be a logistic regression")
    parser.add_argument("--forest", default=False, action="store_true",
                        dest="final_is_forest", help="Final classifier should be a forest")
    parser.add_argument("--voting", default=False, action="store_true",
                        dest="final_is_voting", help="Final classifier should be a voting of all of: forest, svm, logreg")

    options = parser.parse_args()

    if options.fname is None:
        print("ERROR: You must give the pandas file!")
        exit(-1)

    df = pd.read_pickle(options.fname)

    rem_useless_features(df)
    if not options.no_computed:
        add_computed_features(df)

    # cluster setup
    if options.use_clusters:
        used_clusters = df.groupby("clust").nunique()
        clusters = []
        for clust, _ in used_clusters["clust"].iteritems():
            clusters.append(clust)
    else:
        clusters = [0]
        df["clust"] = 0

    # generation
    for clno in clusters:
        funcname = "maple_reward_conf{conf_num}_cluster{clno}".format(
                clno=clno, conf_num=options.conf_num)

        fname = "maple_predictor_conf{conf_num}_cluster{clno}.h".format(
            clno=clno, conf_num=options.conf_num)

        if options.basedir is not None:
            f = options.basedir+"/"+fname
        else:
            f = None

        p = Learner(
            df,
            funcname=funcname,
            fname=f,
            cluster_no=clno)
        p.learn()
